@@ -1,7 +1,7 @@
 # -*- coding: utf-8 -*-
 #
 # gPodder - A media aggregator and podcast client
-# Copyright (c) 2005-2012 Thomas Perl and the gPodder Team
+# Copyright (c) 2005-2016 Thomas Perl and the gPodder Team
 # Copyright (c) 2011 Neal H. Walfield
 #
 # gPodder is free software; you can redistribute it and/or modify
@@ -67,6 +67,28 @@ import feedparser
 import StringIO
 import xml.dom.minidom
 
+import collections
+
+if sys.hexversion < 0x03000000:
+    from HTMLParser import HTMLParser
+    from htmlentitydefs import name2codepoint
+else:
+    from html.parser import HTMLParser
+    from html.entities import name2codepoint
+
+try:
+    import html5lib
+except ImportError:
+    logger.warn('html5lib not found, falling back to HTMLParser')
+    html5lib = None
+
+if gpodder.ui.win32:
+    try:
+        import win32file
+    except ImportError:
+        logger.warn('Running on Win32 but win32api/win32file not installed.')
+        win32file = None
+
 _ = gpodder.gettext
 N_ = gpodder.ngettext
 
@@ -85,9 +107,7 @@ if encoding is None:
         lang = os.environ['LANG']
         (language, encoding) = lang.rsplit('.', 1)
         logger.info('Detected encoding: %s', encoding)
-    elif gpodder.ui.fremantle or gpodder.ui.harmattan:
-        encoding = 'utf-8'
-    elif gpodder.win32:
+    elif gpodder.ui.win32:
         # To quote http://docs.python.org/howto/unicode.html:
         # ,,on Windows, Python uses the name "mbcs" to refer
         #   to whatever the currently configured encoding is``
@@ -132,6 +152,7 @@ _MIME_TYPE_LIST = [
     ('.flv', 'video/x-flv'),
     ('.mkv', 'video/x-matroska'),
     ('.wmv', 'video/x-ms-wmv'),
+    ('.opus', 'audio/opus'),
 ]
 
 _MIME_TYPES = dict((k, v) for v, k in _MIME_TYPE_LIST)
@@ -190,6 +211,12 @@ def normalize_feed_url(url):
 
     >>> normalize_feed_url('http://example.org/test?')
     'http://example.org/test'
+
+    Username and password in the URL must not be affected
+    by URL normalization (see gPodder bug 1942):
+
+    >>> normalize_feed_url('http://UserName:PassWord@Example.com/')
+    'http://UserName:PassWord@example.com/'
     """
     if not url or len(url) < 8:
         return None
@@ -201,7 +228,6 @@ def normalize_feed_url(url):
             'fb:': 'http://feeds.feedburner.com/%s',
             'yt:': 'http://www.youtube.com/rss/user/%s/videos.rss',
             'sc:': 'http://soundcloud.com/%s',
-            'fm4od:': 'http://onapp1.orf.at/webcam/fm4/fod/%s.xspf',
             # YouTube playlists. To get a list of playlists per-user, use:
             # https://gdata.youtube.com/feeds/api/users/<username>/playlists
             'ytpl:': 'http://gdata.youtube.com/feeds/api/playlists/%s',
@@ -218,8 +244,15 @@ def normalize_feed_url(url):
 
     scheme, netloc, path, query, fragment = urlparse.urlsplit(url)
 
+    # Domain name is case insensitive, but username/password is not (bug 1942)
+    if '@' in netloc:
+        authentication, netloc = netloc.rsplit('@', 1)
+        netloc = '@'.join((authentication, netloc.lower()))
+    else:
+        netloc = netloc.lower()
+
     # Schemes and domain names are case insensitive
-    scheme, netloc = scheme.lower(), netloc.lower()
+    scheme = scheme.lower()
 
     # Normalize empty paths to "/"
     if path == '':
@@ -305,6 +338,13 @@ def username_password_from_url(url):
 
     return (username, password)
 
+def directory_is_writable(path):
+    """
+    Returns True if the specified directory exists and is writable
+    by the current user.
+    """
+    return os.path.isdir(path) and os.access(path, os.W_OK)
+
 
 def calculate_size( path):
     """
@@ -371,6 +411,20 @@ def file_age_in_days(filename):
     else:
         return (datetime.datetime.now()-dt).days
 
+def file_modification_timestamp(filename):
+    """
+    Returns the modification date of the specified file as a number
+    or -1 if the modification date cannot be determined.
+    """
+    if filename is None:
+        return -1
+    try:
+        s = os.stat(filename)
+        return s[stat.ST_MTIME]
+    except:
+        logger.warn('Cannot get modification timestamp for %s', filename)
+        return -1
+
 
 def file_age_to_string(days):
     """
@@ -390,6 +444,19 @@ def file_age_to_string(days):
         return N_('%(count)d day ago', '%(count)d days ago', days) % {'count':days}
 
 
+def is_system_file(filename):
+    """
+    Checks to see if the given file is a system file.
+    """
+    if gpodder.ui.win32 and win32file is not None:
+        result = win32file.GetFileAttributes(filename)
+        #-1 is returned by GetFileAttributes when an error occurs
+        #0x4 is the FILE_ATTRIBUTE_SYSTEM constant
+        return result != -1 and result & 0x4 != 0
+    else:
+        return False
+
+
 def get_free_disk_space_win32(path):
     """
     Win32-specific code to determine the free disk space remaining
@@ -397,18 +464,13 @@ def get_free_disk_space_win32(path):
 
     http://mail.python.org/pipermail/python-list/2003-May/203223.html
     """
+    if win32file is None:
+        # Cannot determine free disk space
+        return -1
 
     drive, tail = os.path.splitdrive(path)
-
-    try:
-        import win32file
-        userFree, userTotal, freeOnDisk = win32file.GetDiskFreeSpaceEx(drive)
-        return userFree
-    except ImportError:
-        logger.warn('Running on Win32 but win32api/win32file not installed.')
-
-    # Cannot determine free disk space
-    return 0
+    userFree, userTotal, freeOnDisk = win32file.GetDiskFreeSpaceEx(drive)
+    return userFree
 
 
 def get_free_disk_space(path):
@@ -421,9 +483,9 @@ def get_free_disk_space(path):
     """
 
     if not os.path.exists(path):
-        return 0
+        return -1
 
-    if gpodder.win32:
+    if gpodder.ui.win32:
         return get_free_disk_space_win32(path)
 
     s = os.statvfs(path)
@@ -453,7 +515,10 @@ def format_date(timestamp):
     except ValueError, ve:
         logger.warn('Cannot convert timestamp', exc_info=True)
         return None
-    
+    except TypeError, te:
+        logger.warn('Cannot convert timestamp', exc_info=True)
+        return None
+
     if timestamp_date == today:
        return _('Today')
     elif timestamp_date == yesterday:
@@ -570,6 +635,155 @@ def remove_html_tags(html):
     return result.strip()
 
 
+class HyperlinkExtracter(object):
+    def __init__(self):
+        self.parts = []
+        self.target_stack = [None]
+
+    def get_result(self):
+        # Group together multiple consecutive parts with same link target,
+        # and remove excessive newlines.
+        group_it = itertools.groupby(self.parts, key=lambda x: x[0])
+        result = []
+        for target, parts in group_it:
+            t = u''.join(text for _, text in parts if text is not None)
+            # Remove trailing spaces
+            t = re.sub(' +\n', '\n', t)
+            # Convert more than two newlines to two newlines
+            t = t.replace('\r', '')
+            t = re.sub(r'\n\n\n+', '\n\n', t)
+            result.append((target, t))
+        # Strip leading and trailing whitespace
+        result[0] = (result[0][0], result[0][1].lstrip())
+        result[-1] = (result[-1][0], result[-1][1].rstrip())
+        return result
+
+    def htmlws(self, s):
+        # Replace whitespaces with a single space per HTML spec.
+        if s is not None:
+            return re.sub(r'[ \t\n\r]+', ' ', s)
+
+    def handle_starttag(self, tag_name, attrs):
+        try:
+            handler = getattr(self, 'handle_start_' + tag_name)
+        except AttributeError:
+            pass
+        else:
+            handler(collections.OrderedDict(attrs))
+
+    def handle_endtag(self, tag_name):
+        try:
+            handler = getattr(self, 'handle_end_' + tag_name)
+        except AttributeError:
+            pass
+        else:
+            handler()
+
+    def handle_start_a(self, attrs):
+        self.target_stack.append(attrs.get('href'))
+
+    def handle_end_a(self):
+        if len(self.target_stack) > 1:
+            self.target_stack.pop()
+
+    def output(self, text):
+        self.parts.append((self.target_stack[-1], text))
+
+    def handle_data(self, data):
+        self.output(self.htmlws(data))
+
+    def handle_entityref(self, name):
+        c = unichr(name2codepoint[name])
+        self.output(c)
+
+    def handle_charref(self, name):
+        if name.startswith('x'):
+            c = unichr(int(name[1:], 16))
+        else:
+            c = unichr(int(name))
+        self.output(c)
+
+    def output_newline(self, attrs=None):
+        self.output('\n')
+
+    def output_double_newline(self, attrs=None):
+        self.output('\n')
+
+    def handle_start_img(self, attrs):
+        self.output(self.htmlws(attrs.get('alt', '')))
+
+    def handle_start_li(self, attrs):
+        self.output('\n * ')
+
+    handle_end_li = handle_end_ul = handle_start_br = output_newline
+    handle_start_p = handle_end_p = output_double_newline
+
+
+class ExtractHyperlinkedText(object):
+    def __call__(self, document):
+        self.extracter = HyperlinkExtracter()
+        self.visit(document)
+        return self.extracter.get_result()
+
+    def visit(self, element):
+        NS = '{http://www.w3.org/1999/xhtml}'
+        tag_name = (element.tag[len(NS):] if element.tag.startswith(NS) else element.tag).lower()
+        self.extracter.handle_starttag(tag_name, element.items())
+
+        if element.text is not None:
+            self.extracter.handle_data(element.text)
+
+        for child in element:
+            self.visit(child)
+
+            if child.tail is not None:
+                self.extracter.handle_data(child.tail)
+
+        self.extracter.handle_endtag(tag_name)
+
+
+class ExtractHyperlinkedTextHTMLParser(HTMLParser):
+    def __call__(self, html):
+        self.extracter = HyperlinkExtracter()
+        self.target_stack = [None]
+        self.feed(html)
+        self.close()
+        return self.extracter.get_result()
+
+    def handle_starttag(self, tag, attrs):
+        self.extracter.handle_starttag(tag, attrs)
+
+    def handle_endtag(self, tag):
+        self.extracter.handle_endtag(tag)
+
+    def handle_data(self, data):
+        self.extracter.handle_data(data)
+
+    def handle_entityref(self, name):
+        self.extracter.handle_entityref(name)
+
+    def handle_charref(self, name):
+        self.extracter.handle_charref(name)
+
+
+def extract_hyperlinked_text(html):
+    """
+    Convert HTML to hyperlinked text.
+
+    The output is a list of (target, text) tuples, where target is either a URL
+    or None, and text is a piece of plain text for rendering in a TextView.
+    """
+    if '<' not in html:
+        # Probably plain text. We would remove all the newlines
+        # if we treated it as HTML, so just pass it back as-is.
+        return [(None, html)]
+
+    if html5lib is not None:
+        return ExtractHyperlinkedText()(html5lib.parseFragment(html))
+    else:
+        return ExtractHyperlinkedTextHTMLParser()(html)
+
+
 def wrong_extension(extension):
     """
     Determine if a given extension looks like it's
@@ -648,7 +862,7 @@ def mimetype_from_extension(extension):
     'audio/mpeg'
     >>> mimetype_from_extension('.mkv')
     'video/x-matroska'
-    >>> mimetype_from_extension('.abc')
+    >>> mimetype_from_extension('._invalid_file_extension_')
     ''
     """
     if extension in _MIME_TYPES_EXT:
@@ -786,34 +1000,37 @@ def get_first_line( s):
     return s.strip().split('\n')[0].strip()
 
 
-def object_string_formatter( s, **kwargs):
+def object_string_formatter(s, **kwargs):
     """
-    Makes attributes of object passed in as keyword 
-    arguments available as {OBJECTNAME.ATTRNAME} in 
-    the passed-in string and returns a string with 
-    the above arguments replaced with the attribute 
+    Makes attributes of object passed in as keyword
+    arguments available as {OBJECTNAME.ATTRNAME} in
+    the passed-in string and returns a string with
+    the above arguments replaced with the attribute
     values of the corresponding object.
 
-    Example:
+    >>> class x: pass
+    >>> a = x()
+    >>> a.title = 'Hello world'
+    >>> object_string_formatter('{episode.title}', episode=a)
+    'Hello world'
 
-    e = Episode()
-    e.title = 'Hello'
-    s = '{episode.title} World'
-    
-    print object_string_formatter( s, episode = e)
-          => 'Hello World'
+    >>> class x: pass
+    >>> a = x()
+    >>> a.published = 123
+    >>> object_string_formatter('Hi {episode.published} 456', episode=a)
+    'Hi 123 456'
     """
     result = s
-    for ( key, o ) in kwargs.items():
-        matches = re.findall( r'\{%s\.([^\}]+)\}' % key, s)
+    for key, o in kwargs.iteritems():
+        matches = re.findall(r'\{%s\.([^\}]+)\}' % key, s)
         for attr in matches:
-            if hasattr( o, attr):
+            if hasattr(o, attr):
                 try:
-                    from_s = '{%s.%s}' % ( key, attr )
-                    to_s = getattr( o, attr)
-                    result = result.replace( from_s, to_s)
+                    from_s = '{%s.%s}' % (key, attr)
+                    to_s = str(getattr(o, attr))
+                    result = result.replace(from_s, to_s)
                 except:
-                    logger.warn('Could not replace attribute "%s" in string "%s".', attr, s)
+                    logger.warn('Replace of "%s" failed for "%s".', attr, s)
 
     return result
 
@@ -946,7 +1163,7 @@ def url_add_authentication(url, username, password):
     return urlparse.urlunsplit(url_parts)
 
 
-def urlopen(url):
+def urlopen(url, headers=None, data=None, timeout=None):
     """
     An URL opener with the User-agent set to gPodder (with version)
     """
@@ -960,9 +1177,17 @@ def urlopen(url):
     else:
         opener = urllib2.build_opener()
 
-    headers = {'User-agent': gpodder.user_agent}
-    request = urllib2.Request(url, headers=headers)
-    return opener.open(request)
+    if headers is None:
+        headers = {}
+    else:
+        headers = dict(headers)
+
+    headers.update({'User-agent': gpodder.user_agent})
+    request = urllib2.Request(url, data=data, headers=headers)
+    if timeout is None:
+        return opener.open(request)
+    else:
+        return opener.open(request, timeout=timeout)
 
 def get_real_url(url):
     """
@@ -991,7 +1216,7 @@ def find_command(command):
 
     for path in os.environ['PATH'].split(os.pathsep):
         command_file = os.path.join(path, command)
-        if gpodder.win32 and not os.path.exists(command_file):
+        if gpodder.ui.win32 and not os.path.exists(command_file):
             for extension in ('.bat', '.exe'):
                 cmd = command_file + extension
                 if os.path.isfile(cmd):
@@ -1002,7 +1227,6 @@ def find_command(command):
 
     return None
 
-idle_add_handler = None
 
 def idle_add(func, *args):
     """Run a function in the main GUI thread
@@ -1017,48 +1241,6 @@ def idle_add(func, *args):
     if gpodder.ui.gtk:
         import gobject
         gobject.idle_add(func, *args)
-    elif gpodder.ui.qml:
-        from PySide.QtCore import Signal, QTimer, QThread, Qt, QObject
-
-        class IdleAddHandler(QObject):
-            signal = Signal(object)
-            def __init__(self):
-                QObject.__init__(self)
-
-                self.main_thread_id = QThread.currentThreadId()
-
-                self.signal.connect(self.run_func)
-
-            def run_func(self, func):
-                assert QThread.currentThreadId() == self.main_thread_id, \
-                    ("Running in %s, not %s"
-                     % (str(QThread.currentThreadId()),
-                        str(self.main_thread_id)))
-                func()
-
-            def idle_add(self, func, *args):
-                def doit():
-                    try:
-                        func(*args)
-                    except Exception, e:
-                        logger.exception("Running %s%s: %s",
-                                         func, str(tuple(args)), str(e))
-
-                if QThread.currentThreadId() == self.main_thread_id:
-                    # If we emit the signal in the main thread,
-                    # then the function will be run immediately.
-                    # Instead, use a single shot timer with a 0
-                    # timeout: this will run the function when the
-                    # event loop next iterates.
-                    QTimer.singleShot(0, doit)
-                else:
-                    self.signal.emit(doit)
-
-        global idle_add_handler
-        if idle_add_handler is None:
-            idle_add_handler = IdleAddHandler()
-
-        idle_add_handler.idle_add(func, *args)
     else:
         func(*args)
 
@@ -1115,9 +1297,9 @@ def format_time(value):
     else:
         return dt.strftime('%H:%M:%S')
 
-
 def parse_time(value):
     """Parse a time string into seconds
+
     >>> parse_time('00:00')
     0
     >>> parse_time('00:00:00')
@@ -1130,16 +1312,30 @@ def parse_time(value):
     3600
     >>> parse_time('03:02:01')
     10921
+    >>> parse_time('61:08')
+    3668
+    >>> parse_time('25:03:30')
+    90210
+    >>> parse_time('25:3:30')
+    90210
+    >>> parse_time('61.08')
+    3668
     """
+    if value == '':
+        return 0
+
     if not value:
         raise ValueError('Invalid value: %s' % (str(value),))
 
-    for format in ('%H:%M:%S', '%M:%S'):
-        try:
-            t = time.strptime(value, format)
-            return (t.tm_hour * 60 + t.tm_min) * 60 + t.tm_sec
-        except ValueError, ve:
-            continue
+    m = re.match(r'(\d+)[:.](\d\d?)[:.](\d\d?)', value)
+    if m:
+        hours, minutes, seconds = m.groups()
+        return (int(hours) * 60 + int(minutes)) * 60 + int(seconds)
+
+    m = re.match(r'(\d+)[:.](\d\d?)', value)
+    if m:
+        minutes, seconds = m.groups()
+        return int(minutes) * 60 + int(seconds)
 
     return int(value)
 
@@ -1199,12 +1395,11 @@ def gui_open(filename):
     systems with a few exceptions:
 
        on Win32, os.startfile() is used
-       on Maemo, osso is used to communicate with Nokia Media Player
     """
     try:
-        if gpodder.win32:
+        if gpodder.ui.win32:
             os.startfile(filename)
-        elif gpodder.osx:
+        elif gpodder.ui.osx:
             subprocess.Popen(['open', filename])
         else:
             subprocess.Popen(['xdg-open', filename])
@@ -1220,15 +1415,7 @@ def open_website(url):
     browser. This uses Python's "webbrowser" module, so
     make sure your system is set up correctly.
     """
-    if gpodder.ui.fremantle:
-        import osso
-        context = osso.Context('gPodder', gpodder.__version__, False)
-        rpc = osso.Rpc(context)
-        rpc.rpc_run_with_defaults('osso_browser', \
-                                  'open_new_window', \
-                                  (url,))
-    else:
-        threading.Thread(target=webbrowser.open, args=(url,)).start()
+    run_in_background(lambda: webbrowser.open(url))
 
 def convert_bytes(d):
     """
@@ -1459,9 +1646,7 @@ def detect_device_type():
     Possible return values:
     desktop, laptop, mobile, server, other
     """
-    if gpodder.ui.fremantle or gpodder.ui.harmattan:
-        return 'mobile'
-    elif glob.glob('/proc/acpi/battery/*'):
+    if glob.glob('/proc/acpi/battery/*'):
         # Linux: If we have a battery, assume Laptop
         return 'laptop'
 
@@ -1540,7 +1725,7 @@ def atomic_rename(old_name, new_name):
     the new contents into a temporary file and then moving the
     temporary file over the original file to replace it.
     """
-    if gpodder.win32:
+    if gpodder.ui.win32:
         # Win32 does not support atomic rename with os.rename
         shutil.move(old_name, new_name)
     else:
@@ -1599,3 +1784,117 @@ def get_update_info(url='http://gpodder.org/downloads'):
 
     return up_to_date, latest_version, release_date, days_since_release
 
+
+def run_in_background(function, daemon=False):
+    logger.debug('run_in_background: %s (%s)', function, str(daemon))
+    thread = threading.Thread(target=function)
+    thread.setDaemon(daemon)
+    thread.start()
+    return thread
+
+
+def linux_get_active_interfaces():
+    """Get active network interfaces using 'ip link'
+
+    Returns a list of active network interfaces or an
+    empty list if the device is offline. The loopback
+    interface is not included.
+    """
+    process = subprocess.Popen(['ip', 'link'], stdout=subprocess.PIPE)
+    data, _ = process.communicate()
+    for interface, _ in re.findall(r'\d+: ([^:]+):.*state (UP|UNKNOWN)', data):
+        if interface != 'lo':
+            yield interface
+
+
+def osx_get_active_interfaces():
+    """Get active network interfaces using 'ifconfig'
+
+    Returns a list of active network interfaces or an
+    empty list if the device is offline. The loopback
+    interface is not included.
+    """
+    process = subprocess.Popen(['ifconfig'], stdout=subprocess.PIPE)
+    stdout, _ = process.communicate()
+    for i in re.split('\n(?!\t)', stdout, re.MULTILINE):
+        b = re.match('(\\w+):.*status: (active|associated)$', i, re.MULTILINE | re.DOTALL)
+        if b:
+            yield b.group(1)
+
+def unix_get_active_interfaces():
+    """Get active network interfaces using 'ifconfig'
+
+    Returns a list of active network interfaces or an
+    empty list if the device is offline. The loopback
+    interface is not included.
+    """
+    process = subprocess.Popen(['ifconfig'], stdout=subprocess.PIPE)
+    stdout, _ = process.communicate()
+    for i in re.split('\n(?!\t)', stdout, re.MULTILINE):
+        b = re.match('(\\w+):.*status: active$', i, re.MULTILINE | re.DOTALL)
+        if b:
+            yield b.group(1)
+
+
+def connection_available():
+    """Check if an Internet connection is available
+
+    Returns True if a connection is available (or if there
+    is no way to determine the connection). Returns False
+    if no network interfaces are up (i.e. no connectivity).
+    """
+    try:
+        if gpodder.ui.win32:
+            # FIXME: Implement for Windows
+            return True
+        elif gpodder.ui.osx:
+            return len(list(osx_get_active_interfaces())) > 0
+        else:
+            # By default, we assume we're not offline (bug 1730)
+            offline = False
+
+            if find_command('ifconfig') is not None:
+                # If ifconfig is available, and it says we don't have
+                # any active interfaces, assume we're offline
+                if len(list(unix_get_active_interfaces())) == 0:
+                    offline = True
+
+            # If we assume we're offline, try the "ip" command as fallback
+            if offline and find_command('ip') is not None:
+                if len(list(linux_get_active_interfaces())) == 0:
+                    offline = True
+                else:
+                    offline = False
+
+            return not offline
+
+        return False
+    except Exception, e:
+        logger.warn('Cannot get connection status: %s', e, exc_info=True)
+        # When we can't determine the connection status, act as if we're online (bug 1730)
+        return True
+
+
+def website_reachable(url):
+    """
+    Check if a specific website is available.
+    """
+    if not connection_available():
+        # No network interfaces up - assume website not reachable
+        return (False, None)
+
+    try:
+        response = urllib2.urlopen(url, timeout=1)
+        return (True, response)
+    except urllib2.URLError as err:
+        pass
+
+    return (False, None)
+
+def delete_empty_folders(top):
+    for root, dirs, files in os.walk(top, topdown=False):
+        for name in dirs:
+            dirname = os.path.join(root, name)
+            if not os.listdir(dirname):
+                os.rmdir(dirname)
+
